/*
 * This file implements a file-browser.  The ultimate aim is that, given a
 * wildcard expression, it will display all matching files and will allow
 * you to either pick one, or enter a new name (controlled by a flag).  But
 * for now it simply allows you to enter a new name...
 */

#include "browse.h"
#include <stdio.h>                     /* FILE */
#include <string.h>                    /* strlen(), strcpy() */
#include <stdlib.h>                    /* malloc() */
#ifdef unix
#include <unistd.h>                    /* getcwd() */
#include <glob.h>                      /* glob() */
#else
#include <dir.h>                       /* getcwd() */
#endif

char *browse(char *wildcard, char **filename, char allow_new)
{
  char name[1024];
  char path[1024];
  FILE *the_file;
  int  attrib=0;
#ifdef unix
  glob_t globbuf;
  int    i;
#else
  struct ffblk fcb;
#endif
  char file_not_found = 1;
  
  while (file_not_found) {
    printf("Enter Filename (%s): ", wildcard);
    /* Print out all matching files */
#ifdef unix
    if (glob(wildcard, GLOB_TILDE, NULL, &globbuf)) {
      perror("Error calling glob");
    }
    printf("\n");
    for (i=0; i<globbuf.gl_pathc; i++) {
      printf("%s\n", globbuf.gl_pathv[i]);
    }
    globfree(&globbuf);
#else
    if (findfirst(wildcard, &fcb, attrib) == 0) {
      printf("\n%s\n", fcb.ff_name);
    }
    while (findnext(&fcb)==0) {
      printf("%s\n", fcb.ff_name);
    }
#endif
    /* read in new wildcard / filename */
    fgets(name, 1023, stdin);
    name[strlen(name)-1] = 0;

    /* use wildcard, if no new name was given */
    if (name[0] == 0) {
      strncpy(name, wildcard, 1024);
      name[1023] = 0;
    }
    
    /* allow to exit: */
    /* return NULL, set (*filename)[0]=0 if *filename != NULL */
    
    if (!allow_new) {
      /* check whether file exists! */
      the_file = fopen(name, "r");
      if (the_file == NULL) {
        perror(name);
      } else {
        file_not_found = 0;
      }
    } else {
      file_not_found = 0;
    } /* end new */
  } /* end while */
  
  /* Annotate with current path */
  /* only needed once */
#ifdef unix
  if (name[0] == '/') {
#else
  if (name[1] == ':') {
#endif
    path[0] = 0;
  } else {
    getcwd(path,1024);
    path[1023] = 0;
    path[strlen(path)+1] = 0;
#ifdef unix
    path[strlen(path)]   = '/';
#else
    path[strlen(path)]   = '\\';
#endif
  } /* end else relative path */

  if (*filename == NULL) {
    *filename = malloc(2048);
    if (*filename == NULL) {
      perror("can't allocate memory for filename");
    }
  }
  strncpy(*filename, path, 1024);
  strncat(*filename, name, 2048-2-strlen(*filename));
  return NULL;
}
  
