/* this file describes the file-format for quizcard and defines
 * function to read and write files.  Later this is also going to deal
 * with conversions when reading or writing files.
 */

/* Ok, I think the following file-format might work:
 * 
 * 126 Byte Header:
 *     - 16 Byte Magic Cookie: "@%QUIZCARD 1.0  "
 *     -  2 Byte Codeset Language I
 *     - 30 Byte Language Description Language I
 *     -  2 Byte Codeset Language II
 *     - 30 Byte Language Description Language II
 *     -  2 Byte number of entries
 *     -  1 Byte (Bit) language currently tested
 *     -  1 Byte max_count, 0 means 3 for backwards compatibility
 *     - 43 Byte reserved for future use
 *
 * And for each entry:
 *
 *     -  2 Byte length of entry
 *     -  2 Byte ID
 *     -  1 Byte Byte1
 *     -  1 Byte length of phrase
 *     -  ? Byte Phrase, without the closing zero
 *     -  1 Byte num_translations
 *     -  ?x2 Byte pointers to translations
 *     -  ?x1 Byte count of each translation
 * 
 * It is easy to see from the above that each entry has a length of:
 * (7 + 3*num_translations + length_of(phrase)) Byte
 *
 * The last entry in the file is a 2-byte empty entry, containing 0.
 * This way it is possible to read the size of the following entry
 * together with the current entry, and the size of the first entry
 * together with the header...
 *
 */

extern char max_count;

#include "quizfile.h"
#include "cconvert.h"                  /* cconvert() */
#ifndef unix
#include <io.h>                        /* unlink() */
#endif
#include <stdio.h>                     /* FILE */
#include <string.h>                    /* strrchr() */
#include <errno.h>                     /* errno */
#include <stdlib.h>                    /* realloc() */

char readkey(char* chars);

int write_quiz_file(char  *name,
		    char   lang[2][MAX_PHRASE_LEN],
		    word   the_codeset[2],
		    word   n,
		    byte   current_language_test,
		    entry  card[])
{
  FILE *the_file = NULL;
  int  i,j;
  char header[128];
  int  the_card_size;
  int  max_card_size = 0;
  char *the_card = NULL;
  byte len; /* the length of phrase */
  char *pos;
/*  char new_name[MAX_PHRASE_LEN+1];*/
  char new_name[2][MAX_PHRASE_LEN+1] = {"", ""};

  /* try to open the file to see if it exists */
  the_file = fopen(name, "rb");
  if (the_file != NULL) {
    /* the file exists */
    fclose(the_file);
    /* create names for backup-files */
    strcpy(new_name[0], name);
    strcpy(new_name[1], name);
    pos = strrchr(name,'.');
    if (pos == NULL) {
      strcat(new_name[0], ".#");
      strcat(new_name[1], ".~");
    } else {
      if (strlen(name)-(pos-name)-2 < 3) {
	new_name[0][strlen(new_name[0])-1] = '#';
	new_name[1][strlen(new_name[1])-1] = '~';
      } else {
	strcat(new_name[0], ".#");
	strcat(new_name[1], ".~");
      }
    }
  } else {
    /* no file to take care of... */
    strcpy(new_name[0], name);
    strcpy(new_name[1], "");
  }

  the_file = fopen(new_name[0], "wb");
  if (the_file == NULL) {
    perror(new_name[0]);
    return errno;
  }

  /* compose and write the header */
  for (i=0; i<128; i++) {
    header[i] = 0;
  }
  strncpy(header, "@%QUIZCARD 1.0  ", 16);
  header[16] = the_codeset[0]%256;
  header[17] = the_codeset[0]/256;
  strncpy(&(header[18]), lang[0], 30);
  header[48] = the_codeset[1]%256;
  header[49] = the_codeset[1]/256;
  strncpy(&(header[50]), lang[1], 30);
  header[80] = n%256;
  header[81] = n/256;
  header[82] = current_language_test;
  header[83] = max_count;
  /* we only write the first 126 Byte, as we do not know the size of the
     first entry yet... */
  fwrite(header, 1, 126, the_file);

  /* compose and write each entry */
  for (i=0; i<n; i++) {
    len = strlen(card[i].phrase);
    the_card_size = 7+3*card[i].num_translations+len;
    if (the_card_size > max_card_size) {
      the_card = realloc(the_card, the_card_size);
      if (the_card == NULL) {
				perror("write_quiz_file(): Out of memory");
				exit(1);
      }
      max_card_size = the_card_size;
    }				       /* end cardsize */
    the_card[0] = the_card_size%256;
    the_card[1] = the_card_size/256;
    the_card[2] = card[i].number%256;
    the_card[3] = card[i].number/256;
    the_card[4] = (byte) card[i].byte1;
    the_card[5] = (byte) len;
    strncpy(&(the_card[6]), card[i].phrase, len);
    the_card[6+len] = (byte) card[i].num_translations;
    for (j=0; j<card[i].num_translations; j++) {
      the_card[7+len+3*j+0] = card[i].translations[j]%256;
      the_card[7+len+3*j+1] = card[i].translations[j]/256;
      the_card[7+len+3*j+2] = (byte) card[i].count[j];
    }				       /* end for j */
    fwrite(the_card, 1, the_card_size, the_file);
  }				       /* end for i */

  /* write the last two byte */
  header[0] = 0;
  header[1] = 0;
  fwrite(header, 1, 2, the_file);

  fclose(the_file);

	/* rename files */
	errno = 0;
  /* first, rename an existing file xxx.yyy to xxx.yy~ */
  if (strlen(new_name[1]) != 0) {
#ifndef unix
    /* apparently DOS isn't able to rename over an existing file */
    if (remove(new_name[1])!=0) {
      perror(new_name[1]);
    }
#endif
    if (rename(name, new_name[1])!=0) {
			perror(new_name[1]);
#ifndef unix
			/* apparently DOS isn't able to rename over an existing file */
			if (remove(name)!=0) {
				perror(name);
			}
#endif
		}
    if (rename(new_name[0], name)!=0) {
      perror(name);
    }
		if (errno != 0) {
			readkey("qQ\n\r");
		}
	}				       /* end file exists */

  /* free memory */
  free(the_card);

  return 0;
}

int read_quiz_file(char  *name,
		   char   lang[2][MAX_PHRASE_LEN],
		   word   the_codeset[2],
                   word   to_codeset,
		   word   *n,
		   byte   *current_language_test,
		   entry  card[])
{
  FILE *the_file = NULL;
  int i,j;
  byte header[128];
  word the_card_size;
  word max_card_size = 0;
  byte *the_card = NULL;
  int  resi;
  byte len;
  byte converted = 1;                  /* assume all goes well */
  char* resc;
  char* l_dummy[2] = {NULL};
  char* p_dummy = NULL;

  the_file = fopen(name, "rb");
  if (the_file == NULL) {
    perror(name);
    return errno;
  }

  /* read in the header */
  fread(header, 1, 128, the_file);
  /* check the magick cookie first */
  resi = strncmp((char*)header, "@%QUIZCARD 1.0  ", 16);
  if (resi != 0) {
    header[16] = 0; /* terminate String */
    fprintf(stderr, "Magick Cookie should be:\n%s\nis\n%s\n",
	    "@%QUIZCARD 1.0  ", header);
    exit(1);
  }
  the_codeset[0] = (word) (header[16] + 256*header[17]);
  the_codeset[1] = (word) (header[48] + 256*header[49]);
  if ((converted) && (the_codeset[0] != to_codeset)) {
    resc = cconvert((char*)&(header[18]),
                    the_codeset[0],
                    &(l_dummy[0]),
                    to_codeset);
    if (resc == NULL) {
      converted = 0;
    }
    strncpy(lang[0], l_dummy[0], 30);
    free(l_dummy[0]);
  } else {
    strncpy(lang[0], (char*)&(header[18]), 30);
  }
  lang[0][30] = 0;
  if ((converted) && (the_codeset[1] != to_codeset)) {
    resc = cconvert((char*)&(header[50]),
                    the_codeset[1],
                    &(l_dummy[1]),
                    to_codeset);
    if (resc == NULL) {
      converted = 0;
    }
    strncpy(lang[1], l_dummy[1], 30);
    free(l_dummy[1]);
  } else {
    strncpy(lang[1], (char*)&(header[50]), 30);
    lang[1][30] = 0;
  }
  *n = (word) (header[80] + 256*header[81]);
  *current_language_test = (byte) header[82];
  max_count = (byte) header[83];
  /* backwards compatibility */
  if (max_count == 0) {
    max_count = 3;
  }
  the_card_size = (word) (header[126] + 256*header[127]);
  fprintf(stderr, "Languages in header: %s and %s\n", &(header[18]), &(header[50]));
  fprintf(stderr, "Language 0: %s, Codeset = %d\n", lang[0], the_codeset[0]);
  fprintf(stderr, "Language 1: %s, Codeset = %d\n", lang[1], the_codeset[1]);
  fprintf(stderr, "Reading %d entries, first entry is %d bytes\n", *n, the_card_size);

  for (i=0; i<*n; i++) {
    if (the_card_size > max_card_size) {
      the_card = realloc(the_card, the_card_size);
      if (the_card == NULL) {
	perror("read_quiz_file()");
	exit(1);
      }
      max_card_size = the_card_size;
    }				       /* end if not enough mem */
    fread(the_card, 1, the_card_size, the_file);
    card[i].number = (word) (the_card[0] + 256*the_card[1]);
    if (card[i].number != i) {
      fprintf(stderr, "Wrong card-number: should be %d, is %d!\n", i, card[i].number);
    }
    card[i].byte1 = (byte) the_card[2];
    len = ((byte) the_card[3]);
    p_dummy = malloc(len+1);
    if (p_dummy == NULL) {
      perror("allocating memory for phrase");
      fprintf(stderr, "Could not allocate %d bytes\n",len+1);
      exit(1);
    }
    strncpy(p_dummy, (char*)&(the_card[4]), len);
    p_dummy[len] = 0;
    if ((converted) && (the_codeset[(card[i].byte1&LANGUAGE)>>7] != to_codeset)) {
      resc = cconvert(p_dummy,
                     the_codeset[(card[i].byte1&LANGUAGE)>>7],
                     &card[i].phrase,  /* should be either NULL or allocated! */
                     to_codeset);
      if (resc == NULL) {
        converted = 0;
      }
      free(p_dummy);
    } else {
      free(card[i].phrase);	       /* should be either NULL or allocated! */
      card[i].phrase = p_dummy;
    }
    p_dummy = NULL;
    /* fprintf(stderr, "phrase    = %s\n", card[i].phrase); */
		card[i].num_translations   = the_card[4+len];
    /* fprintf(stderr, "num_trans = %d\n", card[i].num_translations); */
    free(card[i].translations);	       /* should be either NULL or allocated! */
    free(card[i].count);	       /* should be either NULL or allocated! */
    card[i].translations = malloc(2*card[i].num_translations);
    card[i].count        = malloc(card[i].num_translations);
    if ((card[i].translations != 0) && ((card[i].translations == NULL) || (card[i].count == NULL))) {
      perror("allocating memory for translations");
      fprintf(stderr, "Could not allocate %d*3 Byte.\n",
	      card[i].num_translations);
      exit(1);
    }
    for (j=0; j<card[i].num_translations; j++) {
      card[i].translations[j] = (word) (the_card[5+len+3*j]   + 256*the_card[5+len+3*j+1]);
      card[i].count[j]        = (byte)  the_card[5+len+3*j+2];
      if (card[i].count[j] > max_count) {
	card[i].count[j] = max_count;
      }
    }				       /* end for j */
    the_card_size = (word) (the_card[the_card_size-2] + 256*the_card[the_card_size-2+1]);
  }				       /* end for i */

  if (converted) {
    the_codeset[0] = to_codeset;
    the_codeset[1] = to_codeset;
  }
  
  /* free memory */
  free(the_card);

  return 0;
}
