#include "test.h"
#include "quizcard.h"
#include "sort.h"		       /* show_*(), sprint_phrase() */
#include "cconvert.h"		       /* recode() */
#include <string.h>		       /* strncpy() */
#include <stdio.h>		       /* fprintf() */
#include <stdlib.h>		       /* rand() */
/* system-specific stuff.  This should go into a header-file of it's
   own, and possibly even into the makefile ... */
#ifdef unix
#include "conio.h"
#else
#include <conio.h>
#endif

extern char max_count;

/* This function selects a card for testing which hadn't previously 
 * been drawn in this session. */

int random_draw(entry cards[MAX_NUM_ENTRIES],
		word n,		       /* number of cards */
		byte l		       /* current language */
		)
{
  int i;
  double dummy;
  
  do {
    i = rand();
    dummy = ((double)n)/(RAND_MAX+1.0);
    i = i*dummy;
  } while ((cards[i].byte1 & TESTED) ||
	   (((cards[i].byte1 & LANGUAGE) >> 7) != l));
  return i;
}

/*
 * This function implements the actual comparison.  At first it will only
 * do a simple string-comparison, but later on it should do more complicated
 * things, e.g. ignore whitespace differences, ignore text inside braces
 * (and the braces) or deal with choices as signified by a slash (/).
 */


int is_equal(char* s1,
	     char* s2)
{
  char ds[2][1024];
  int i1, i2, j;
  int res;

	strncpy(ds[0], s1, 1023);
  ds[0][1023] = 0;
  strncpy(ds[1], s2, 1023);
  ds[1][1023] = 0;

  for (j=0; j<2; j++) {
    /* remove text delimited by [] or () */
    for (i1=0, i2=0; i2<strlen(ds[j]); ) {
      switch (ds[j][i2]) {
      case '[':
        do {
          i2++;
        } while (ds[j][i2] != ']');
        i2++;
        break;
      case '(':
        do {
          i2++;
        } while (ds[j][i2] != ')');
        i2++;
        break;
      default:
        ds[j][i1] = ds[j][i2];
        i1++;
        i2++;
        break;
      }
    } /* end for i */
    ds[j][i1] = 0;
    /* remove spaces at the beginning */
    i2=0;
    while (ds[j][i2] == ' ') {
      i2++;
    }
    /* remove double (or more) spaces */
    for (i1=0; i2<strlen(ds[j]); ) {
      switch (ds[j][i2]) {
      case ' ':
	ds[j][i1++] = ds[j][i2++];
        while (ds[j][i2] == ' ') {
          i2++;
        }
	switch (ds[j][i2]) {
	 case '.':
	 case ',':
	 case ';':
	 case ':':
	 case '!':
	 case '?':
	  /* no space before punctuation mark */
	 case 0:
          /* no space at end of string */
	  ds[j][i1-1] = ds[j][i2];
	  break;
        }
        i2++;
        break;
      default:
        ds[j][i1] = ds[j][i2];
        i1++;
        i2++;
        break;
      }
    } /* end for i */
    ds[j][i1] = 0;
  } /* end for j */
  
  /* 0 means is equal */
  res = ! strcmp(ds[0],ds[1]);
  
  return res;
}


/*
 * 
 * This function implements the actual test-routine.  The basic idea is
 * a) vocabulary training is done in sessions.  During each session _all_
 *    entries (which are not already considered learned) are tested exactly
 *    once.  If a correct translation is given, the score for _this_ 
 *    combination is increased (all other possible translations are left 
 *    unchanged).  If an incorrect translation was given, than _all_
 *    translations are set back to zero.
 * b) once an answer has been given, the following is displayed:
 *    * An indication of failure or success (failure can also be signalled
 *      audibly)
 *    * all correct translations.
 *    * Iff an incorrect answer was given: the translation of that (wrong)
 *      answer, iff one exists.
 * c) testing can be stopped and resumed any time.
 */
int test(entry cards[MAX_NUM_ENTRIES],
	 word  n[2],		       /* number of entries */
	 word  n_tested[2],	       /* number of tested entries */
	 word  n_pos[2],	       /* of cards known */
	 word  n_neg[2],	       /* number of failed cards */
	 char  lang[2][MAX_PHRASE_LEN],
	 word  the_coding_system,
	 byte  *current_language_test
	 )
{
  int    i, j;			       /* loop control variables */
  char   answer[MAX_PHRASE_LEN+1]="";  /* answer-string */
  char   c = 'n';		       /* answer-character */
  word   l;			       /* current language */
  char   the_phrase[MAX_PHRASE_LEN+1];
  int    equal;			       /* flag */
  int    tested;		       /* flag */
  word   correct;

  l = *current_language_test;

  do {
    if (n_tested[l] == n[l]) {
      clrscr();
      gotoxy(32,9);
      printf("Congratulation!\n");
      if ((n_pos[l]+n_neg[l]) > 1) {
	gotoxy(19,11);
	printf("%d of %d (%4.1f%%) cards sucessfully tested!\n", n_pos[l], n_pos[l]+n_neg[l], n_pos[l]*100.0/(n_pos[l]+n_neg[l]));
      } else {
	gotoxy(31,11);
	printf("All cards tested!\n");
      }
      gotoxy(26,14);
      printf("Press RETURN to continue...");
      n_tested[l] = 0;
      n_pos[l]    = 0;
      n_neg[l]    = 0;
      for (i=0; i<n[2]; i++) {
	cards[i].byte1 = cards[i].byte1 & (255-TESTED);
	/* have any cards passed the final test? */
	tested = 1;
	for (j=0; j<cards[i].num_translations; j++) {
	  if (cards[i].count[j] < max_count) {
	    tested = 0;
	  }			       /* end if */
	}			       /* end for j */
	if (tested) {
	  cards[i].byte1 = cards[i].byte1 | TESTED;
	  n_tested[(cards[i].byte1 & LANGUAGE) >> 7]++;
	}
      }				       /* end for i */
      *current_language_test = (*current_language_test+1)%2;
      readkey("qQ\r\n");
      return 0;
    }				       /* end if n */
#ifdef blabla
    do {
      /* 0 <= i < n */
      i = rand();
      dummy = ((double)n[2])/((double)RAND_MAX+1);
      i = i*dummy;
      /* i = i*(((double)n)/((double)(RAND_MAX+1)));*/
    } while ((cards[i].byte1 & TESTED) ||
	     (((cards[i].byte1 & LANGUAGE) >> 7) != l));
#endif
    i = random_draw(cards, n[2], l);
    clrscr();
    printf("%d tested, %d remaining", n_tested[l], n[l]-n_tested[l]);
    if ((n_pos[l]+n_neg[l]) > 0) {
			printf(", %4.1f%% correct\n", n_pos[l]*100.0/(n_pos[l]+n_neg[l]));
    } else {
      printf("\n");
    }
    gotoxy(1,3);
    printf("%14s", lang[l]);
    gotoxy(15,3);
    printf(" : %s", sprint_phrase(the_phrase, &(cards[i])));
    for (j=0; j<cards[i].num_translations; j++) {
      printf(" %d", cards[i].count[j]);
    }				       /* end for all translations */
    gotoxy(1,5);
    printf("%14s : ", lang[(l+1)%2]);
    fflush(stdout);
    fgets(answer, MAX_PHRASE_LEN, stdin);
    answer[strlen(answer)-1] = 0;
    /* reparse the answer to swap specially encoded characters */
    recode(answer, MAX_PHRASE_LEN+1, the_coding_system);

    cards[i].byte1 = cards[i].byte1 | TESTED;
    n_tested[l]++;
    /* test against all possible translations */
    equal = 0;
    for (j=0; j<cards[i].num_translations; j++) {
      if (is_equal(cards[cards[i].translations[j]].phrase, answer)) {
	equal = 1;
	correct = cards[i].translations[j];
	if (cards[i].count[j] < max_count) {
	  cards[i].count[j]++;
	} else {
	  cards[i].count[j] = max_count;
	}
      }			       /* end is equal */
    }			       /* end for all translations */
    if (!equal) {
      /* print correct translations */
      clrscr();
      printf("INCORRECT answer: %s\n", answer);
      printf("Correct answers:\n");
      show_translations(cards, i);
      gotoxy(50,13);
      printf("n) next   card");
      gotoxy(50,14);
      printf("d) delete card");
      gotoxy(50,15);
      printf("e) edit   card");
      gotoxy(50,16);
      printf("i) ignore card");
      gotoxy(50,18);
      printf("q) quit testing");
      gotoxy(1,1);
      fflush(stdout);
      for (j=0; j<cards[i].num_translations; j++) {
	if (cards[i].count[j] < max_count) {
	  cards[i].count[j] = 0;
	} else {
	  cards[i].count[j] = max_count;
	}
      }			       /* end for all translations */
      n_neg[l]++;
      c = readkey("dDeEiInNqQ\r\n");
      switch (c) {
       case 'd':
       case 'D':
       case 'e':
       case 'E':
	/* not yet implemented */
	break;
       case 'i':
       case 'I':
	for (j=0; j<cards[i].num_translations; j++) {
	  cards[i].count[j] = max_count;
	}			       /* end for all translations */
	break;
       case 'n':
       case 'N':
       case '\r':
       case '\n':
	/* Next Card */
	c = 'n';
	break;
       case 'q':
       case 'Q':
	/* Quit testing */
	c = 'q';
	break;
      }
    } else {
      n_pos[l]++;
      if (cards[i].num_translations > 1) {
	/* was equal, but maybe there are additional translations? */
	for (j=0; j<cards[i].num_translations; j++) {
	  if ((cards[i].count[j] < max_count) && (cards[i].translations[j] != correct)) {
	    clrscr();
	    printf("\n");
	    printf("Other correct answers:\n\n");
	    show_alternatives(cards, i,correct,20);
	    printf("\n");
	    printf("\nPress 'q' when ready\n");
	    j = cards[i].num_translations;
	    readkey("qQ\r\n");
	  }			       /* end not learned */
	}			       /* end for all translations */
      }				       /* end if several translations */
    }				       /* end is equal */
  } while (c != 'q');

  return 0;
}
